import org.apache.commons.lang3.StringUtils

import com.google.common.collect.LinkedListMultimap
import org.apache.poi.ss.util.CellReference

import com.midori.cloud.renderer.jira.model.component.ColumnLayoutItem
import com.midori.cloud.renderer.jira.model.component.IssueTableLayoutBean

// check dependency
if (!binding.variables.containsKey('fieldHelper')) {
	throw new Exception("The script \"issue-navigator.groovy\" requires the prior execution of \"field-helper-tool.groovy\".\nUse this expression in one of the topmost cells: <mt:execute script=\"field-helper-tool.groovy\"/><mt:execute script=\"issue-navigator.groovy\"/>")
}

// set these explicitly to avoid "not a property" exceptions
if (!binding.variables.containsKey('tempo')) {
	tempo = null
}

if (searchRequest?.name != null) {
	def templateExtension = StringUtils.substringAfterLast(xlsView?.templateName, '.')
	filename = "${searchRequest.name}.${templateExtension ?: 'xlsx'}".toString() // "filename" is a standard Java String, not a GString
}

issueNavigator = new IssueNavigatorTool(componentAccessor, xlsView, user, searchRequest, fieldHelper, tempo)

/**
 * Tool to export dynamic column configurations from Issue Navigator.
 * It requires the prior execution of "field-helper-tool.groovy".
 */
class IssueNavigatorTool {
	/**
	 * Comma-separated list of the custom field type keys to exclude from the export.
	 */
	def EXCLUDED_CUSTOM_FIELD_TYPE_KEYS = ["com.pyxis.greenhopper.jira:gh-epic-color", "com.pyxis.greenhopper.jira:gh-global-rank", "com.pyxis.greenhopper.jira:gh-lexo-rank", "com.atlassian.jira.ext.charting:firstresponsedate", "com.atlassian.jira.ext.charting:timeinstatus"]
	/**
	 * Comma-separated list of the custom field IDs to exclude from the export (e.g.: [10001, 10015]).
	 */
	def EXCLUDED_CUSTOM_FIELD_IDS = []

	/**
	 * Comma-separated list of the field IDs to export to the leftmost columns in "all fields" and "non-empty fields" modes (e.g.: ["issuekey", "summary", "customfield_10000"])
 	 */
	def LEFTMOST_COLUMNS = ["issuekey", "summary"]

	def xlsView
	def user
	def searchRequest
	def fieldHelper
	def i18nHelper
	def tableLayoutFactory
	def tempo

	def issueTableLayout

	public IssueNavigatorTool(def componentAccessor, def xlsView, def user, def searchRequest, def fieldHelper, def tempo) {
		this.xlsView = xlsView
		this.user = user
		this.searchRequest = searchRequest
		this.fieldHelper = fieldHelper
		this.i18nHelper = componentAccessor.i18nHelper
		this.tableLayoutFactory = componentAccessor.tableLayoutFactory
		this.tempo = tempo
	}

	/**
	 * Returns whether all fields should be exported.
	 * @see https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/recipes#all-vs-non-empty-vs-current-fields
	 */
	private isAllFieldsMode() {
		viewNameContains("all")
	}

	/**
	 * Returns whether non-empty fields should be exported.
	 * @see https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/recipes#all-vs-non-empty-vs-current-fields
	 */
	private isNonEmptyFieldsMode() {
		viewNameContains("non-empty") || viewNameContains("non-e.")
	}

	/**
	 * Returns whether current fields should be exported.
	 * @see https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/recipes#all-vs-non-empty-vs-current-fields
	 */
	private isCurrentFieldsMode() {
		!isAllFieldsMode() && !isNonEmptyFieldsMode()
	}

	private viewNameContains(s) {
		return (boolean) xlsView?.name?.toLowerCase()?.contains(s)
	}

	/**
	 * @deprecated
	 * This method does not support the "non-empty" mode because it doesn't receive the issues.
	 * Use {@link #getIssueTableLayout(java.lang.Object)} instead.
	 */
	@Deprecated
	def getIssueTableLayout() {
		getIssueTableLayout(null)
	}

	def getIssueTableLayout(issues) {
		// lazily calculate the table layout
		if (issueTableLayout == null) {
			def isCurrentFieldsMode = isCurrentFieldsMode()
			def originalIssueTableLayout = isCurrentFieldsMode ? tableLayoutFactory.getStandardExcelLayout(searchRequest, user) : tableLayoutFactory.getAllColumnsExcelLayout()

			// filter and sort fields
			def position = 0
			def filteredColumns = originalIssueTableLayout.columns.findAll {
				def columnLayoutItem = it
				def field = columnLayoutItem.navigableField
				def isIncludedByValue = (issues && isNonEmptyFieldsMode()) ? isColumnNonEmpty(columnLayoutItem, issues) : true
				if (field.id.startsWith("customfield_")) {
					def isExcludedByType = EXCLUDED_CUSTOM_FIELD_TYPE_KEYS.contains(field.customFieldType.key)
					def isExcludedById = EXCLUDED_CUSTOM_FIELD_IDS.contains(field.idAsLong.intValue())
					return (!isExcludedByType && !isExcludedById && isIncludedByValue)
				} else {
					return isIncludedByValue
				}
			}.sort({ a, b ->
				if (isCurrentFieldsMode) {
					return 0
				}
				def indexA = LEFTMOST_COLUMNS.indexOf(a.navigableField.id)
				def indexB = LEFTMOST_COLUMNS.indexOf(b.navigableField.id)
				if ((indexA > -1) || (indexB > -1)) {
					indexA = (indexA > -1) ? indexA : LEFTMOST_COLUMNS.size()
					indexB = (indexB > -1) ? indexB : LEFTMOST_COLUMNS.size()
					return indexA <=> indexB
				}
				def headingA = i18nHelper.getText(a.navigableField.columnHeadingKey).toLowerCase()
				def headingB = i18nHelper.getText(b.navigableField.columnHeadingKey).toLowerCase()
				return headingA <=> headingB
			}).collect { it ->
				new ColumnLayoutItem(it.navigableField, position++)
			}

			issueTableLayout = new IssueTableLayoutBean(filteredColumns)
		}

		return issueTableLayout
	}

	/**
	 * Returns whether at least one issue has non-empty value in the passed column.
	 */
	private isColumnNonEmpty(columnLayoutItem, issues) {
		issues.any { issue ->
			def value = getFieldValue(issue, columnLayoutItem)
			return value && !String.valueOf(value).endsWith("not supported yet")
		}
	}

	def getFieldValue(issue, columnLayoutItem) {
		fieldHelper.getFieldValue(issue, columnLayoutItem.navigableField)
	}

	def getFieldCellType(columnLayoutItem) {
		fieldHelper.getFieldCellType(columnLayoutItem.navigableField)
	}

	def getFieldFormat(columnLayoutItem) {
		fieldHelper.getFieldFormat(columnLayoutItem.navigableField)
	}

	/**
	 * Returns the "aggregation" formula to be inserted into the last row in the given column or <code>null</code>
	 * if the column should not be aggregated. This is typically a formula using SUM(), COUNTA() or such.
	 */
	def getFieldSumFormula(columnLayoutItem, cell) {
		def cellReference = new CellReference(cell)
		def firstDataRowCellReferenceString = new CellReference(1, cell.columnIndex).formatAsString()
		def lastDataRowCellReferenceString = new CellReference(cell.rowIndex - 1, cell.columnIndex).formatAsString()

		!fieldHelper.isCustomField(columnLayoutItem.navigableField) ? getSystemFieldSumFormula(columnLayoutItem.navigableField, firstDataRowCellReferenceString, lastDataRowCellReferenceString) : getCustomFieldSumFormula(columnLayoutItem.navigableField, firstDataRowCellReferenceString, lastDataRowCellReferenceString)
	}

	private getSystemFieldSumFormula(field, firstDataRowCellReferenceString, lastDataRowCellReferenceString) {
		switch(field.id) {
			case "issuekey":
				// count of issues
				"COUNTA(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString})"
				break

			case "timeoriginalestimate":
			case "timeestimate":
			case "timespent":
			case "aggregatetimespent":
			case "aggregatetimeestimate":
			case "aggregatetimeoriginalestimate":
			case "votes":
			case "watches":
				// sum of values
				"SUM(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString})"
				break

			default:
				// no formula
				null
		}

	}

	private getCustomFieldSumFormula(customField, firstDataRowCellReferenceString, lastDataRowCellReferenceString) {
		switch(customField.customFieldType.key) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:float":
			case "com.atlassian.jira.toolkit:attachments":
			case "com.atlassian.jira.toolkit:comments":
			case "com.pyxis.greenhopper.jira:jsw-story-points":
				// sum of values
				"SUM(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString})"
				break

			// Jira Service Management custom field types
			case "com.atlassian.servicedesk:sd-request-feedback":
			case "com.atlassian.servicedesk:sd-sla-field":
				// average of values
				"ROUND(AVERAGE(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString}),2)"
				break

			default:
				// no formula
				null
		}
	}

	/**
	 * Returns a "repeating" list of entries where the issue is the repeating key and its worklog is the value.
	 * Example: (i1 -> w11, i1 -> w12, i2 -> w21, ...)
	 */
	def getRepeatingWorklogs(issues) {
		def result = LinkedListMultimap.create() // keeps insertion order for both keys and values

		issues.each {
			it ->
				def worklogs = it.worklogsRaw ?: [[:]]
				result.putAll(it, worklogs)
		}

		return result.entries()
	}

	/**
	 * Returns a "repeating" list of entries where the issue is the repeating key and its Tempo worklog is the value.
	 * Example: (i1 -> tw11, i1 -> tw12, i2 -> tw21, ...)
	 */
	def getRepeatingTempoWorklogs(issues) {
		def result = LinkedListMultimap.create() // keeps insertion order for both keys and values

		issues.each {
			it ->
				def tempoWorklogs = tempo.getWorklogs(it.key) ?: [[:]]
				result.putAll(it, tempoWorklogs)
		}

		return result.entries()
	}

	/**
	 * Returns a "repeating" list of entries where the (issue, change history) pair is the repeating key and the change item is the value.
	 * Example: (i1 -> ci111, i1 -> ci112, i1 -> c121, i2 -> ci211, ...)
	 */
	def getRepeatingChangeHistories(issues) {
		def result = LinkedListMultimap.create() // keeps insertion order for both keys and values

		issues.each { issue ->
			def changeHistories = issue.changeHistories ?: [[:]]
			changeHistories.each { changeHistory ->
				def changeItems = changeHistory.changeItemBeans ?: [[:]]
				changeItems.each { changeItem ->
					def changeHistoryAndChangeItem = [ authorDisplayName: changeHistory.authorDisplayName, timePerformed: changeHistory.timePerformed, field: changeItem.field, fromString: changeItem.fromString, toString: changeItem.toString ]
					result.put(issue, changeHistoryAndChangeItem)
				}
			}
		}

		return result.entries()
	}
}
