filename = "${title} - ${new Date().format('yyyy-MM-dd')}.xlsx".toString() // "filename" is a standard Java String, not a GString

timeInStatusReport = new TimeInStatusReportTool()

/**
 * Tool to help with workflow analysis.
 */
class TimeInStatusReportTool {
	/**
	 * @return the list of maps (entries for "name", "start", "end" and "running") that represent the time metrics.
	 */
	def getTimeMetrics(def issue) {
		// get time metrics for past transitions
		def metrics = []
		def now = new Date()
		def start = issue.getCreated()

		for (changeHistory in issue.changeHistories) {
			for (changeItem in changeHistory.changeItems) {
				if ("status".equals(changeItem.field)) {
					def m = [name: "Time in \"${changeItem.oldstring}\" (h)", start: start, end: changeHistory.timePerformed, "running": false]
					metrics.add(m)

					start = changeHistory.timePerformed
				}
			}
		}

		// get time metric for current status
		metrics.add([name: "Time in \"${issue.statusObject.nameTranslation}\" (h)", start: start, end: now, "running": true])

		// get calculated time metrics
		def firstResponseTime = getIssueFirstResponseTime(issue)
		metrics.add([name: "Time to First Response (h)", start: issue.getCreated(), end: firstResponseTime ? firstResponseTime : now, "running": (firstResponseTime == null)])

		def resolutionTime = getIssueResolutionTime(issue)
		metrics.add([name: "Time to Resolution (h)", start: issue.getCreated(), end: resolutionTime ? resolutionTime : now, "running": (resolutionTime == null)])

		return metrics
	}

	/**
	 * @return a list of the status names in which the issue has ever been in chronological order
	 * (e.g. [ "To Do", "In Progress", "To Do", "In Progress", "Done" ]).
	 */
	def getHistoricalStatusNames(def issue) {
		def statusNames = []

		// add historical statuses
		for (changeHistory in issue.changeHistories) {
			for (changeItem in changeHistory.changeItems) {
				if ('status'.equals(changeItem.field)) {
					statusNames.add(changeItem.oldstring)
				}
			}
		}

		// add current status
		statusNames.add(issue.statusObject.nameTranslation)

		return statusNames
	}

	/**
	 * Returns the creation time of the latest comment or <code>null</code> if the issue has no comments.
	 */
	private static def getIssueFirstResponseTime(issue) {
		def comments = issue.comments
		return !comments.isEmpty() ? comments[0].created : null
	}

	/**
	 * Returns the resolution time or <code>null</code> for unresolved issues.
	 */
	private static def getIssueResolutionTime(issue) {
		issue.resolutionDate
	}
}